
/* Copyright  2014 Pierre Vidal <p.terkal@gmail.com>
 * This program is free software. It comes without no warranty.
 * You can redistribute it and/or modify it under the terms of the
 * Do What The Fuck You Want To Public License, Version 2, as published by Sam Hocevar.
 * See the COPYING file or http://www.wtfpl.net/ for more details. */

#include "main.h"

#include <stdio.h>
#include <string.h>

struct mech_weapon {
    char* name;
    int cooldown;
    float x_speed;
    float y_speed;
    enum missile_type missile;
};

struct mech_thruster {
    char* name;
    float speed;
    SDL_Color const * trail_color;
    int trail_width;
    int trail_length;
};

struct mech_special {
    char* name;
    int cooldown;
    int delay;
    int duration;
    enum special_actif_type type;
};

struct mech_gear {
    int nb_weapon;
    struct mech_weapon weapon[MAX_GEAR];
    int nb_thruster;
    struct mech_thruster thruster[MAX_GEAR];
    int nb_special;
    struct mech_special special[MAX_GEAR];
};

char* path(char* subdata_rep, char* filename) {
    static char path[BUFSIZ];
    if(subdata_rep == NULL) {
        return(filename);
    } else if(subdata_rep[0] == '\0') { // == ""
#ifdef SHMUP_WINDOWS
        sprintf(path, "data\\%s", filename);
#else
        sprintf(path, "data/%s", filename);
#endif
    } else {
#ifdef SHMUP_WINDOWS
        sprintf(path, "data\\%s\\%s", subdata_rep, filename);
#else
        sprintf(path, "data/%s/%s", subdata_rep, filename);
#endif
    }
    return(path);
}

char* path_localized(char* subdata_rep, char* filename) {
    static char path[BUFSIZ];
    if(subdata_rep == NULL) {
        return(filename);
    } else if(subdata_rep[0] == '\0') { // == ""
#ifdef SHMUP_WINDOWS
        sprintf(path, "data\\%s\\%s", local.code, filename);
#else
        sprintf(path, "data/%s/%s", local.code, filename);
#endif
    } else {
#ifdef SHMUP_WINDOWS
        sprintf(path, "data\\%s\\%s\\%s", subdata_rep, local.code, filename);
#else
        sprintf(path, "data/%s/%s/%s", subdata_rep, local.code, filename);
#endif
    }
    return(path);
}

char* copy(char* src) {
    char* dst = malloc(strlen(src)+1);
    if(dst == NULL)
        error("not enough memory"), exit(EXIT_FAILURE);
    strcpy(dst, src);
    return(dst);
}

void add_wave(struct run* run, int spawn_per_sec, int model_id, int timestart, int duration) {
    if(run->nb_waves < MAX_WAVE-1) {
        run->waves[run->nb_waves].model_id = model_id;
        run->waves[run->nb_waves].timestart = timestart * 100;
        run->waves[run->nb_waves].duration = duration * 100;
        run->waves[run->nb_waves].spawn_per_sec = spawn_per_sec;
        run->waves[run->nb_waves].accumulator = 0;
        run->nb_waves++;
    }
}

void add_dialogue(SDL_Renderer* renderer, struct run* run, char* text, int timestart, int duration, enum speaker speaker, int y_shift) {
    if(run->nb_dialogue < MAX_DIALOGUE-1) {
        SDL_Color const * color;
        switch(speaker) {
            case speaker_muship : color = color_muship; break;
            case speaker_player : color = color_player; break;
            case speaker_mech : color = color_mech; break;
            case speaker_evil : color = color_evil; break;
            case speaker_nhy : color = color_nhy; break;
        }
        run->dialogues[run->nb_dialogue].text = text;
        run->dialogues[run->nb_dialogue].timestart = timestart*100;
        run->dialogues[run->nb_dialogue].duration = duration*100;
        run->dialogues[run->nb_dialogue].speaker = speaker;
        run->dialogues[run->nb_dialogue].y_shift = y_shift * lineskip;
        run->dialogues[run->nb_dialogue].texture =
            loadTextureText(renderer, text, &(run->dialogues[run->nb_dialogue].w), &(run->dialogues[run->nb_dialogue].h), color);
        run->nb_dialogue++;
    }
}

enum speaker a2speaker(char* str) {
    if(!strcmp(str, "muship")) {
        return(speaker_muship);
    } else if(!strcmp(str, "player")) {
        return(speaker_player);
    } else if(!strcmp(str, "mech")) {
        return(speaker_mech);
    } else if(!strcmp(str, "evil")) {
        return(speaker_evil);
    } else if(!strcmp(str, "nhy")) {
        return(speaker_nhy);
    } else {
        return(-1);
    }
}

int a2ship(char* str, struct ship_model ship_models[], int nb_models) {
    for(int i=0; i<nb_models; i++) {
        if(!strcmp(str, ship_models[i].name)) {
            return(i);
        }
    }
    return(-1);
}

void load_level(SDL_Renderer* renderer, struct run* run, int run_id, struct ship_model ship_models[], int nb_ships) {
    FILE* file;
    char* filename;
    char buf[BUFSIZ];
    char* token[6];
    int line = 0;

    run->nb_dialogue = 0;
    run->nb_waves = 0;

    if(run_id == ARCADE_LEVEL) {
        filename = copy(path("scripts", "arcade.txt"));
    } else {
        sprintf(buf, "level_%02d.txt", run_id);
        filename = copy(path_localized("levels", buf));
    }

    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open level file '%s'", filename);
        error(buf);
        return;
    }

    while(fgets(buf, BUFSIZ/*-1*/, file) != NULL) {
        int i=0;
        line++;

        buf[strlen(buf)-1] = '\0'; // Cut trailing '\n'.
        if(buf[0] == '#' || buf[0] == '\0')
            continue;

        if(buf[0] <= '9' && buf[0] >= '0') { // Add wave.
            token[i] = strtok(buf, " ");
            while(i < 4) {
                if(token[i] == NULL) {
                    fprintf(stderr, "WARNING: file '%s', line %d: wave has not enough parameters.\n", filename, line);
                    break;
                }
                token[++i] = strtok(NULL, " ");
            }
            if(i == 4) {
                int ship_id = a2ship(token[1], ship_models, nb_ships);
                if(ship_id == -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized mechip '%s'.\n", filename, line, token[1]);
                } else {
                    add_wave(run, atoi(token[0]), ship_id, atoi(token[2]), atoi(token[3]));
                }
            }

        } else if(buf[0] == '"') { // Add dialogue.
            token[i] = strtok(buf, "\"");
            while(i < 5) {
                if(token[i] == NULL) {
                    fprintf(stderr, "WARNING: file '%s', line %d: dialogue has not enough parameters.\n", filename, line);
                    break;
                }
                token[++i] = strtok(NULL, " ");
            }
            if(i==5) {
                enum speaker spkr = a2speaker(token[3]);
                if(spkr == -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized speaker '%s'.\n", filename, line, token[3]);
                } else {
                    add_dialogue(renderer, run, token[0], atoi(token[1]), atoi(token[2]), spkr, atoi(token[4]));
                }
            }
        }
    }

    free(filename);
    fclose(file);
}

void add_mechip(SDL_Renderer* renderer, struct ship_model ship_model[], int* nb_ships, char* name, char* texture_file_name, int hp) {
    if(*nb_ships < MAX_SHIP_MODEL-1) {
        ship_model[*nb_ships].name = copy(name);
        ship_model[*nb_ships].texture = loadTexture(renderer, path("ships", texture_file_name));
        ship_model[*nb_ships].gear.hp = hp;
        ship_model[*nb_ships].gear.nb_thruster = 0;
        ship_model[*nb_ships].gear.nb_weapon = 0;
        ship_model[*nb_ships].gear.nb_special_actif = 0;
        ship_model[*nb_ships].gear.nb_special_passif = 0;
        ship_model[*nb_ships].gear.delta_v = 0;
        ship_model[*nb_ships].gear.speed = 0;
        if(ship_model[*nb_ships].texture != NULL) {
            (*nb_ships)++;
        } else {
            free(ship_model[*nb_ships].name);
        }
    } else {
        fprintf(stderr, "Too many ships defined.\n");
    }
}

SDL_Color const * const a2trail(char* str) {
    if(!strcmp(str, "helltrail")) {
        return(color_helltrail);
    } else if(!strcmp(str, "mushtrail")) {
        return(color_mushtrail);
    } else if(!strcmp(str, "darkmushtrail")) {
        return(color_darkmushtrail);
    } else {
        return(NULL);
    }
}

enum missile_type a2missile(char* str) {
    if(!strcmp(str, "mechile")) {
        return(missile_mechile);
    } else if(!strcmp(str, "mechmine")) {
        return(missile_mechmine);
    } else if(!strcmp(str, "mechbullet")) {
        return(missile_mechbullet);
    } else if(!strcmp(str, "mechautobullet")) {
        return(missile_mechautobullet);
    } else {
        return(-1);
    }
};

enum special_actif_type a2special(char* str) {
    if(!strcmp(str, "beamlazer")) {
        return(special_actif_beamlazer);
    } else {
        return(-1);
    }
}

int a2weapon_id(struct mech_gear * mgear, char* str) {
    for(int i=0; i<mgear->nb_weapon; i++) {
        if(!strcmp(mgear->weapon[i].name, str)) {
            return(i);
        }
    }
    return(-1);
}
int a2thruster_id(struct mech_gear * mgear, char* str) {
    for(int i=0; i<mgear->nb_thruster; i++) {
        if(!strcmp(mgear->thruster[i].name, str)) {
            return(i);
        }
    }
    return(-1);
}

int a2special_id(struct mech_gear * mgear, char* str) {
    for(int i=0; i<mgear->nb_special; i++) {
        if(!strcmp(mgear->special[i].name, str)) {
            return(i);
        }
    }
    return(-1);
}

void load_mech_gear(struct mech_gear * mgear) {
    FILE* file;
    int line = 0;
    char* filename;
    char buf[BUFSIZ];
    char* token[6];

    mgear->nb_weapon = 0;
    mgear->nb_thruster = 0;
    mgear->nb_special = 0;

    filename = copy(path("scripts", "mech_gear.txt"));
    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open mech's gear file '%s'", filename);
        error(buf);
        return;
    }

    while(fgets(buf, BUFSIZ/*-1*/, file) != NULL) {
        int i;
        line++;

        buf[strlen(buf)-1] = '\0'; // Cut trailing '\n'.
        if(buf[0] == '#' || buf[0] == '\0')
            continue;

        token[0] = strtok(buf, " ");
        for(i=1; i<6; i++) {
            token[i] = strtok(NULL, " ");
            if(token[i] == NULL) {
                fprintf(stderr, "WARNING: file '%s', line %d: not enough parameters.\n", filename, line);
                break;
            }
        }
        if(i == 6) {
            if(token[0][0] == 'W') { // Weapon.
                if(mgear->nb_weapon >= MAX_GEAR-1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: too much weapons.\n", filename, line);
                } else if(a2weapon_id(mgear, token[1]) != -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: weapon '%s' already exists.\n", filename, line, token[1]);
                } else if(i = a2missile(token[5]), i == -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized type of missile '%s'.\n", filename, line, token[5]);
                } else {
                    mgear->weapon[mgear->nb_weapon].name = copy(token[1]);
                    mgear->weapon[mgear->nb_weapon].cooldown = atoi(token[2]);
                    mgear->weapon[mgear->nb_weapon].x_speed = atof(token[3]);
                    mgear->weapon[mgear->nb_weapon].y_speed = atof(token[4]);
                    mgear->weapon[mgear->nb_weapon].missile = i;
                    mgear->nb_weapon++;
                }
            } else if(token[0][0] == 'T') { // Thruster.
                if(mgear->nb_thruster >= MAX_GEAR-1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: too much thrusters.\n", filename, line);
                } else if(a2thruster_id(mgear, token[1]) != -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: thruster '%s' already exists.\n", filename, line, token[1]);
                } else if(i = (int) a2trail(token[3]), i == -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized type of trail '%s'.\n", filename, line, token[3]);
                } else {
                    mgear->thruster[mgear->nb_thruster].name = copy(token[1]);
                    mgear->thruster[mgear->nb_thruster].speed = atof(token[2]);
                    mgear->thruster[mgear->nb_thruster].trail_color = (SDL_Color const * const) i;
                    mgear->thruster[mgear->nb_thruster].trail_width = atoi(token[4]);
                    mgear->thruster[mgear->nb_thruster].trail_length = atoi(token[5]);
                    mgear->nb_thruster++;
                }
            } else if(token[0][0] == 'S') { // Special.
                if(mgear->nb_special >= MAX_GEAR-1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: too much special weapons.\n", filename, line);
                } else if(a2special_id(mgear, token[1]) != -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: special weapon '%s' already exists.\n", filename, line, token[1]);
                } else if(i = a2special(token[5]), i == -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized type of special weapon '%s'.\n", filename, line, token[5]);
                } else {
                    mgear->special[mgear->nb_special].name = copy(token[1]);
                    mgear->special[mgear->nb_special].cooldown = atoi(token[2]);
                    mgear->special[mgear->nb_special].delay = atoi(token[3]);
                    mgear->special[mgear->nb_special].duration = atoi(token[4]);
                    mgear->special[mgear->nb_special].type = i;
                    mgear->nb_special++;
                }
            } else {
                fprintf(stderr, "WARNING: file '%s', line %d: unrecognized gear category '%s'.\n", filename, line, token[0]);
            }
        }
    }

    free(filename);
    fclose(file);
}

void unload_mechgear(struct mech_gear * mgear) {
    for(int i=0; i<mgear->nb_weapon; i++) {
        free(mgear->weapon[i].name);
    }
    for(int i=0; i<mgear->nb_thruster; i++) {
        free(mgear->thruster[i].name);
    }
    for(int i=0; i<mgear->nb_special; i++) {
        free(mgear->special[i].name);
    }
}

void load_mechip_models(SDL_Renderer* renderer, struct ship_model ship_model[], int* nb_ships) {
    FILE* file;
    char buf[BUFSIZ];
    int line = 0;
    char* filename;
    char* token[5];

    struct mech_gear mgear;
    load_mech_gear(&mgear);

    *nb_ships = 0;

    filename = copy(path("scripts", "mechip.txt"));
    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open mech's ships file '%s'", filename);
        error(buf);
        return;
    }

    while(fgets(buf, BUFSIZ/*-1*/, file) != NULL) {
        line++;
        int i;
        int is_ship = true;

        buf[strlen(buf)-1] = '\0'; // Cut trailing '\n'.
        if(buf[0] == '#' || buf[0] == '\0')
            continue;

        if(buf[0] == '\t')
            is_ship = false;

        token[0] = strtok(is_ship ? buf : buf+1, " ");
        if(is_ship) { // Add ship.
            for(i=1; i<3; i++) {
                token[i] = strtok(NULL, " ");
                if(token[i] == NULL) {
                    fprintf(stderr, "WARNING: file '%s', line %d: not enough parameters.\n", filename, line);
                    break;
                }
            }
            if(i == 3) {
                if(a2ship(token[0], ship_model, *nb_ships) != -1) {
                    fprintf(stderr, "WARNING: file '%s', line %d: ship '%s' already exists.\n", filename, line, token[0]);
                } else {
                    add_mechip(renderer, ship_model, nb_ships, token[0], token[1], atoi(token[2]));
                }
            }
        } else { // Add gear.
            for(i=1; i<4; i++) {
                token[i] = strtok(NULL, " ");
                if(token[i] == NULL) {
                    fprintf(stderr, "WARNING: file '%s', line %d: not enough parameters.\n", filename, line);
                    break;
                }
            }
            if(*nb_ships == 0) {
                fprintf(stderr, "WARNING: file '%s', line %d: a gear must be declared after a ship.\n", filename, line);
            } else if(i == 4) {
                token[4] = strtok(NULL, " "); // Get optional delay.
                if(token[0][0] == 'W') { // Weapon.
                    i = a2weapon_id(&mgear, token[1]);
                    if(i == -1) {
                        fprintf(stderr, "WARNING: file '%s', line %d: unrecognized weapon '%s'.\n", filename, line, token[1]);
                    } else {
                        add_weapon(&(ship_model[*nb_ships-1].gear), atoi(token[2]), atoi(token[3]), mgear.weapon[i].cooldown,
                                   token[4] == NULL ? 0 : atoi(token[4]), // Optional delay.
                                   mgear.weapon[i].x_speed, mgear.weapon[i].y_speed, mgear.weapon[i].missile);
                    }
                } else if(token[0][0] == 'T') { // Thruster.
                    i = a2thruster_id(&mgear, token[1]);
                    if(i == -1) {
                        fprintf(stderr, "WARNING: file '%s', line %d: unrecognized thruster '%s'.\n", filename, line, token[1]);
                    } else {
                        add_thruster(&(ship_model[*nb_ships-1].gear), atoi(token[2]), atoi(token[3]),
                                     mgear.thruster[i].trail_width, mgear.thruster[i].trail_length, mgear.thruster[i].trail_color);
                        ship_model[*nb_ships-1].gear.speed += mgear.thruster[i].speed;
                    }
                } else if(token[0][0] == 'S') { // Special.
                    i = a2special_id(&mgear, token[1]);
                    if(i == -1) {
                        fprintf(stderr, "WARNING: file '%s', line %d: unrecognized thruster '%s'.\n", filename, line, token[1]);
                    } else {
                        add_special_actif(&(ship_model[*nb_ships-1].gear), atoi(token[2]), atoi(token[3]),
                                          mgear.special[i].cooldown,
                                          (token[4] == NULL ? 0 : atoi(token[4])) + mgear.special[i].delay, // Optional delay.
                                          mgear.special[i].duration,
                                          mgear.special[i].type);
                    }
                } else {
                    fprintf(stderr, "WARNING: file '%s', line %d: unrecognized gear category '%s'.\n", filename, line, token[0]);
                }
            }
        }
    }

    unload_mechgear(&mgear);
    free(filename);
    fclose(file);
}

void unload_mechip_models(int nb_ships, struct ship_model ship_model[]) {
    for(int i=0; i<nb_ships; i++) {
        free(ship_model[i].name);
        SDL_DestroyTexture(ship_model[i].texture);
    }
}

void load_briefing(char* txt[], int chapter) {
    char buf[BUFSIZ];
    char* filename;
    FILE* file;
    int i = 0;

    txt[0] = ""; // '\0';
    sprintf(buf, "briefing_%02d.txt", chapter);
    filename = copy(path_localized("levels", buf));
    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open briefing file '%s'", filename);
        error(buf);
        return;
    }

    while(fgets(buf, BUFSIZ/*-1*/, file) != NULL) {
        buf[strlen(buf)-1] = '\0'; // Cut trailing '\n'.
        if(buf[0] == '\0') {
            txt[i] = copy(" ");
        } else {
            txt[i] = copy(buf);
        }
        i++;
        if(i >= MAX_BRIEFING-1) {
            break;
        }
    }
    txt[i] = ""; // '\0';
}

void unload_briefing(char* txt[]) {
    for(int i=0; txt[i][0] != '\0'; i++) {
        free(txt[i]);
    }
}

int load_cinematic_text(int chapter, char* txt[], int nb_txt) {
    char buf[BUFSIZ];
    char* filename;
    FILE* file;
    int i = 0;

    txt[0] = ""; // '\0';
    sprintf(buf, "cinematic_%02d.txt", chapter);
    filename = copy(path_localized("levels", buf));
    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open cinematic text file '%s'", filename);
        error(buf);
        return(EXIT_FAILURE);
    }

    while(fgets(buf, BUFSIZ/*-1*/, file) != NULL) {
        buf[strlen(buf)-1] = '\0'; // Cut trailing '\n'.
        if(buf[0] == '\0') {
            txt[i] = copy(" ");
        } else {
            txt[i] = copy(buf);
        }
        i++;
        if(i >= MAX_BRIEFING-1) {
            break;
        }
    }

    return(EXIT_SUCCESS);
}

void unload_cinematic_text(char* txt[], int nb_txt) {
    for(int i=0; i < nb_txt && txt[i][0] != '\0'; i++) {
        free(txt[i]);
    }
}

void load_local(char* code) {
    FILE* file;
    char* filename;
    char buf[BUFSIZ];
    strncpy(local.code, code, 2);

    filename = path_localized("locale", "menu_options.txt");
    file = fopen(filename, "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<option_entry_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.optionmenu_entry[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "menu_sound.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<sound_entry_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.soundmenu_entry[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "menu_keybind_hints.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<MAX_KEYBIND_HINT; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.keybindmenu_hints[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "menu_keybind_entry.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<game_keybind_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.keybindmenu_entry[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "gene_console.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_changekeybindF1 = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_resetkeybindF1 = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_buildyourship = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_doublevalid = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_doubleescape = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_weight = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_power = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_agility = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_speed = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_shield = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_ram = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "gene_invalid_messages.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<invalid_reason_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_invalidmsg[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "gene_keybind_entry.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<gene_keybind_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_keybind[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "gene_keybind_hint.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else for(int i=0; i<gene_keybind_max; i++) {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.console_hint[i] = copy(buf);
    }
    fclose(file);

    file = fopen(path_localized("locale", "menu_display.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.optimenu_enable = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.optimenu_disable = copy(buf);
        for(int i=0; i<display_entry_max; i++) {
            fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
            local.optimenu_entry[i] = copy(buf);
        }
    }
    fclose(file);

    file = fopen(path_localized("locale", "score.txt"), "r");
    if(file == NULL) {
        sprintf(buf, "Unable to open locale file '%s'", filename);
        error(buf);
    } else {
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_pressenter = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_ult_geneticefficiency = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_geneticefficiency = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_ult_mushroomavenger = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_mushroomavenger = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_impossible = copy(buf);
        fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
        local.gameover_accuracy = copy(buf);
        for(int i=0; i<display_entry_max; i++) {
            fgets(buf, BUFSIZ, file), buf[strlen(buf)-1] = '\0';
            local.gameover_scores[i] = copy(buf);
        }
    }
    fclose(file);
}

void unload_local() {
    for(int i=0; i<option_entry_max; i++) {
        free(local.optionmenu_entry[i]);
    }
    free(local.storymenu_returntomain);
    for(int i=0; i<MAX_KEYBIND_HINT; i++) {
        free(local.keybindmenu_hints[i]);
    }
    for(int i=0; i<game_keybind_max; i++) {
        free(local.keybindmenu_entry[i]);
    }
    for(int i=0; i<display_entry_max; i++) {
        free(local.optimenu_entry[i]);
    }
    free(local.optimenu_enable);
    free(local.optimenu_disable);
    free(local.gameover_pressenter);
    for(int i=0; i<score_entry_max; i++) {
        free(local.gameover_scores[i]);
    }
    free(local.gameover_accuracy);
    free(local.gameover_ult_geneticefficiency);
    free(local.gameover_geneticefficiency);
    free(local.gameover_ult_mushroomavenger);
    free(local.gameover_mushroomavenger);
    free(local.gameover_impossible);
    for(int i=0; i<gene_keybind_max; i++) {
        free(local.console_hint[i]);
        free(local.console_keybind[i]);
    }
    free(local.console_changekeybindF1);
    free(local.console_resetkeybindF1);
 //   char* console_partname[part_type_max];
    free(local.console_buildyourship);
    free(local.console_doublevalid);
    free(local.console_doubleescape);
    for(int i=0; i<invalid_reason_max; i++) {
        free(local.console_invalidmsg[i]);
    }
    free(local.console_weight);
    free(local.console_power);
    free(local.console_agility);
    free(local.console_speed);
    free(local.console_shield);
    free(local.console_ram);
}
