
/* Copyright  2014 Pierre Vidal <p.terkal@gmail.com>
 * This program is free software. It comes without no warranty.
 * You can redistribute it and/or modify it under the terms of the
 * Do What The Fuck You Want To Public License, Version 2, as published by Sam Hocevar.
 * See the COPYING file or http://www.wtfpl.net/ for more details. */

#include "main.h"

#include <math.h>

#define SHIP_SIZE           32
#define PARTICLE_SIZE       5
#define MAX_MISSILE         256
#define MAX_SHIP            128
#define MAX_EXPLOSION       64
#define MAX_PARTICLE        30
#define MAX_PARTICLE_SPEED  10
#define PARTICLE_SPEED_RATIO 0.1
#define EXPLOSION_PHASE_MAX 4
#define EXPLOSION_DURATION  40
#define ENDGAME_DELAY       150
#define LAZER_WIDTH         10
#define LAZER_HALO_RADIUS   20
#define FRONTSPAWN_DELAY    7
#define FRONTSPAWN_VARIANCE 4
#define EXPLOSION_SOUND_VOLUME 0.5

struct missile {
    float x;
    float y;
    float dx;
    float dy;
    float dy_shift;
    float speed; // from 0 to 1
    enum missile_type type;
    int alliance;
#define ALLIANCE_MUSH 0
#define ALLIANCE_MECH 1
};

struct missile_model {
    SDL_Texture* texture;
    int w;
    int h;
    int dammage;
    float acceleration;
    SDL_Color const * color_trail; // NULL for no trail.
};

struct ship {
    float x;
    float y;
    int cd[MAX_GEAR]; // Current cooldown for each gear->weapon[].
    int special_cd[MAX_GEAR];
    int hp; // Current hit points ; not max.
    int type; // ship id
    int skirt; // bool. Is doppler thruster sound already played?
};

struct explosion {
    float x;
    float y;
    int remains;
    int type;
#define EXPLOSION_TYPE_MECH 0
#define EXPLOSION_TYPE_MUSH 1
#define EXPLOSION_TYPE_MAX  2
};

struct particle {
    float x;
    float y;
    float speed;
    int type;
};

enum particle_type {
    particle_blue,
    particle_green,
    particle_violet,
    particle_red,
    particle_type_max
};

// Return the phase of the explosion (0~3) from its remaining time.
int phase(int remains) {
    return(
            remains < EXPLOSION_DURATION/4 ? 3 :
            remains < EXPLOSION_DURATION/2 ? 2 :
            remains < EXPLOSION_DURATION/4*3 ? 1 : 0
    );
}

void add_ship(struct ship* ship, int type, float x, float y, struct ship_model* ship_model) {
    ship->type = type;
    ship->x = x;
    ship->y = y;
    ship->hp = ship_model[type].gear.hp;
    ship->skirt = false;
    for(int i=0; i<ship_model[type].gear.nb_weapon; i++)
        ship->cd[i] = ship_model[type].gear.weapon[i].cd / 2;
    for(int i=0; i<ship_model[type].gear.nb_special_actif; i++) {
        ship->special_cd[i] = rand() % (ship_model[type].gear.special_actif[i].cd / 4) + ship_model[type].gear.special_actif[i].cd / 4;
    }
}

void add_missile(struct missile* missile, struct weapon* weapon, float ship_x, float ship_y, int alliance, float delta_y) {
    missile->type = weapon->missile;
    missile->x = ship_x + weapon->x;
    missile->y = ship_y + weapon->y;
    missile->dx = weapon->xspeed;
    missile->dy = weapon->yspeed;
    missile->dy_shift =  delta_y;
    missile->speed = 0.0;
    missile->alliance = alliance;
    if(weapon->missile == missile_mechmine || weapon->missile == missile_mushmine) {
        missile->dy_shift = 0;
    }
}

void add_explosion(struct explosion* expl, int* nb_expl, float x, float y, int type) {
    if(*nb_expl < MAX_EXPLOSION-1) {
        expl[*nb_expl].x = x;
        expl[*nb_expl].y = y;
        expl[*nb_expl].remains = EXPLOSION_DURATION;
        expl[*nb_expl].type = type;
        if(!mute) {
            // add_sound(type == EXPLOSION_TYPE_MECH ? sound_explosion : sound_explosion_grave, 0.3);
            add_sound(sound_explosion, 0.3);
// TODO : grave explosion sound by muship explosion musn't be ovewritten. How?
// TODO : set volume function of distance between it and the muship.
        }

        (*nb_expl)++;
    }
}

int /*bool*/ collision(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2) {
    if(abs(y1-y2) <= (h1+h2)/2) { // Check Y position first
        if(abs(x1-x2) <= (w1+w2)/2) {
            float dx2, dy2;
            dx2 = (x1-x2); dx2 *= dx2;
            dy2 = (y1-y2); dy2 *= dy2;
            if(sqrt(dx2+dy2) < (w1+w2)/2) { // TODO : use ellipse collision.
                return(true);
            }
            // TODO : avoid divide W and H by 2 in pre-check?
        }
    }
    return(false);
}

void build_missile_models(SDL_Renderer* renderer, struct missile_model model[]) {
    char* filename;
    for(enum missile_type i=0; i<missile_type_max; i++) {
        model[i].dammage = 1;
        model[i].w = model[i].h = 8;
        model[i].acceleration = 1;
        model[i].color_trail = NULL;
        switch(i) {
            case missile_mechile:
                filename = path("missile", "mechile.bmp");
                model[i].acceleration = 0.05;
                model[i].color_trail = color_helltrail;
                break;
            case missile_mushile:
                filename = path("missile", "mushile.bmp");
                model[i].acceleration = 0.1;
                model[i].color_trail = color_mushtrail;
                break;
            case missile_mechbullet:
            case missile_mechautobullet:
                filename = path("missile", "mechbullet.bmp");
                break;
            case missile_mushbullet:
                filename = path("missile", "mushbullet.bmp");
                break;
            case missile_mechmine:
                filename = path("missile", "mechmine.bmp");
                model[i].w = model[i].h = 12;
                break;
            case missile_mushmine:
                filename = path("missile", "mushmine.bmp");
                model[i].w = model[i].h = 12;
                break;
            case missile_sniper:
                filename = path("missile", "mushile.bmp");
                model[i].acceleration = 1;
                model[i].dammage = 4;
                model[i].color_trail = color_mushtrail;
                break;
            case missile_torpedo:
                filename = path("missile", "mushile.bmp");
                model[i].acceleration = 0.01;
                model[i].dammage = 4;
                model[i].color_trail = color_mushtrail;
                break;
            case missile_type_max: break;
        }
        model[i].texture = loadTexture(renderer, filename);
    }
}

void draw_trail(SDL_Renderer* renderer, int x, int y, int W, int H, SDL_Color const * const color) {
    SDL_SetRenderDrawColor(renderer, color->r, color->g, color->b, color->a);
    for(int h=0; abs(h)<abs(H); H>0 ? h++ : h--) {
        float max_w = abs(H-h)*W/abs(H);
        for(int w=0; w < max_w; w++) {
            if(rand()%abs(H) >= abs(h)) {
                SDL_RenderDrawPoint(renderer, x - (max_w/2) + w, y+h);
            }
        }
    }
}

int game(SDL_Renderer* renderer, int run_id, SDL_Keycode keybind[game_keybind_max]) {
    char* filename;
    int stop = false;   // Main loop's stop flag.
    int lose = 0;       // Number of frame since muship destroyed.
    int win  = 0;       // Number of frame since last item clean.
    int result = -1;    // True = win, false = lose, -1 for error or exit.
    int frontspawn;     // Ship spawned since last "front spawn".
    long time;          // Frames since the begining of the game. Long int should be enough...
    long last_tick;     // Used for frame limit.
    SDL_Rect dst;       // It's always useful to have a rectangle at hand!
    int lftkey, rgtkey, upkey, dwnkey, firekey, spekey;
                        // Booleans storing if key is pressed.

    struct run run;
    struct score score;

    SDL_Texture* the_muship_texture;
    int the_muship_w;
    int the_muship_h;
    float the_muship_x;
    float the_muship_y;
    struct gear the_muship_gear;
    int the_muship_cd[MAX_GEAR];
    int the_muship_special_cd[MAX_GEAR];

    SDL_Texture* icon_shield;
    SDL_Rect icon_shield_dst;

    struct ship_model ship_models[MAX_SHIP_MODEL]; // List of ship's models.
    int nb_ship_models;
    struct ship ships[MAX_SHIP]; // List of ships on the screen.
    int nb_ships;

    struct missile_model missile_models[missile_type_max];
    struct missile missiles[MAX_MISSILE]; // List of missiles on the screen.
    int nb_missiles;

    SDL_Texture* particle_texture[particle_type_max];
    struct particle particles[MAX_PARTICLE]; // List of particles on the screen.

    SDL_Texture* explosion_texture[EXPLOSION_PHASE_MAX][EXPLOSION_TYPE_MAX];
    struct explosion explosions[MAX_EXPLOSION]; // List of explosions on the screen.
    int nb_explosions;

    // Load textures.

    icon_shield = loadTextureSize(renderer, path("", "icon_shield.bmp"), &(icon_shield_dst.w), &(icon_shield_dst.h));
    if(opti.explosion) {
        explosion_texture[0][EXPLOSION_TYPE_MECH] = loadTexture(renderer, path("", "explode0.bmp"));
        explosion_texture[1][EXPLOSION_TYPE_MECH] = loadTexture(renderer, path("", "explode1.bmp"));
        explosion_texture[2][EXPLOSION_TYPE_MECH] = loadTexture(renderer, path("", "explode2.bmp"));
        explosion_texture[3][EXPLOSION_TYPE_MECH] = loadTexture(renderer, path("", "explode3.bmp"));
        explosion_texture[0][EXPLOSION_TYPE_MUSH] = loadTexture(renderer, path("", "explode0_m.bmp"));
        explosion_texture[1][EXPLOSION_TYPE_MUSH] = loadTexture(renderer, path("", "explode1_m.bmp"));
        explosion_texture[2][EXPLOSION_TYPE_MUSH] = loadTexture(renderer, path("", "explode2_m.bmp"));
        explosion_texture[3][EXPLOSION_TYPE_MUSH] = loadTexture(renderer, path("", "explode3_m.bmp"));
    }
    if(opti.particles) for(enum particle_type i=0; i<particle_type_max; i++) {
        switch(i) {
            case particle_blue: filename = path("", "particle_blue.bmp"); break;
            case particle_green: filename = path("", "particle_green.bmp"); break;
            case particle_violet: filename = path("", "particle_violet.bmp"); break;
            case particle_red: filename = path("", "particle_red.bmp"); break;
            case particle_type_max: break;
        }
        particle_texture[i] = loadTexture(renderer, filename);
    }

    build_missile_models(renderer, missile_models);
    // nb_ship_models = 0; // load_mechip_model() already do it.
    load_mechip_models(renderer, ship_models, &nb_ship_models);
    load_level(renderer, &run, run_id, ship_models, nb_ship_models);

    // Build the muship.
    {
        SDL_Surface* surface = build_archip(renderer, &the_muship_gear, run_id);
        if(surface == NULL)
            return(-1);
        the_muship_texture = SDL_CreateTextureFromSurface(renderer, surface);
        the_muship_w = surface->w;
        the_muship_h = surface->h;
        SDL_FreeSurface(surface);
    }

    // Check texture loading.
    if(icon_shield == NULL) {
        return(-1);
    }
    for(int i=0; i<missile_type_max; i++) {
        if(missile_models[i].texture == NULL) {
            return(-1);
        }
    }
    for(int i=0; i<nb_ship_models; i++) {
        if(ship_models[i].texture == NULL) {
            return(-1);
        }
    }
    if(opti.explosion) for(int i=0; i<EXPLOSION_PHASE_MAX; i++) {
        for(int j=0; j<EXPLOSION_TYPE_MAX; j++) {
            if(explosion_texture[i][j] == NULL) {
                return(-1);
            }
        }
    }
    if(opti.particles) for(int i=0; i<particle_type_max; i++) {
        if(particle_texture[i] == NULL) {
            return(-1);
        }
    }

    // Init game state.

    lftkey = rgtkey = upkey = dwnkey = firekey = spekey = 0;
    nb_missiles = nb_explosions = nb_ships = 0;
    time = 0;
    last_tick = SDL_GetTicks();
    frontspawn = FRONTSPAWN_DELAY;

    if(!mute) {
        begin_sound();
    }

    if(opti.particles) for(int i=0; i<MAX_PARTICLE; i++) {
        particles[i].type = rand() % particle_type_max;
        particles[i].speed = ((rand()%MAX_PARTICLE_SPEED)+1)*PARTICLE_SPEED_RATIO;
        particles[i].x = rand() % SCREEN_W;
        particles[i].y = rand() % SCREEN_H;
    }

    for(enum score_entry i=0; i<score_entry_max; i++) {
        switch(i) {
            case score_destroy: score.nb_destroy = 0; break;
            case score_dodge: score.nb_dodge = 0; break;
            case score_shot: score.nb_shot = 0; break;
            case score_hit: score.nb_hit = 0; break;
            case score_time: score.time = 0; break;
            case score_entry_max: break;
        }
    }

    the_muship_x = SCREEN_W / 2;
    the_muship_y = SCREEN_H - 100;
    for(int i=0; i<the_muship_gear.nb_weapon; i++) {
        // the_muship_cd[i] = the_muship_gear.weapon[i].cd/2;
        the_muship_cd[i] = 0;
    }
    for(int i=0; i<the_muship_gear.nb_special_actif; i++) {
        // the_muship_special_cd[i] = the_muship_gear.special_actif[i].cd/2;
        the_muship_special_cd[i] = 0;
    }

    while(!stop) {
        SDL_Event event;

        // Get keyboard inputs.

        while(SDL_PollEvent(&event)) {
            if(event.type == SDL_QUIT || ( event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE ) ) {
                stop = true;
            } else if(event.type == SDL_KEYDOWN || event.type == SDL_KEYUP) {
                int* ptr;
                if(event.key.keysym.sym == keybind[game_keybind_moveup]) {
                    ptr = &upkey;
                } else if(event.key.keysym.sym == keybind[game_keybind_movedown]) {
                    ptr = &dwnkey;
                } else if(event.key.keysym.sym == keybind[game_keybind_moveleft]) {
                    ptr = &lftkey;
                } else if(event.key.keysym.sym == keybind[game_keybind_moveright]) {
                    ptr = &rgtkey;
                } else if(event.key.keysym.sym == keybind[game_keybind_fire]) {
                    ptr = &firekey;
                } else if(event.key.keysym.sym == keybind[game_keybind_altfire]) {
                    ptr = &spekey;
                } else {
                        ptr = NULL;
                }
                if(ptr != NULL) {
                    if(event.type == SDL_KEYDOWN) {
                        *ptr = 1;
                    } else { // SDL_KEYUP
                        *ptr = 0;
                    }
                }
            }
        }

        // Update model.

        // Update muship's coordinates.
        if(!lose) {
            int x=0;
            int y=0;
            // TODO : better diagonals calculation
            if(lftkey)
                x--;
            if(rgtkey)
                x++;
            if(upkey)
                y--;
            if(dwnkey)
                y++;

            if(x && y) {
                // cos(45) == sqrt(2)/2 =~ 0.70XX
                the_muship_x += x * the_muship_gear.delta_v * 0.7 * 5;
                the_muship_y += y * the_muship_gear.delta_v * 0.7 * 5;
            } else if(x) {
                the_muship_x += x * the_muship_gear.delta_v * 5;
            } else if(y) {
                the_muship_y += y * the_muship_gear.delta_v * 5;
            }

            // The muship bump against screen's edges.
            if(the_muship_x < 0 + the_muship_w/2) {
                the_muship_x = the_muship_w/2;
            } else if(the_muship_x > SCREEN_W - the_muship_w/2) {
                the_muship_x = SCREEN_W - the_muship_w/2;
            }

            if(the_muship_y < 0 + the_muship_h/2) {
                the_muship_y = the_muship_h/2;
            } else if(the_muship_y > SCREEN_H - the_muship_h/2) {
                the_muship_y = SCREEN_H - the_muship_h/2;
            }
        }

        // Update mechips' coordinates and destroy lost ones.
        for(int i=nb_ships-1; i>=0; i--) {
            ships[i].y += ship_models[ships[i].type].gear.speed + the_muship_gear.speed;
            if(ships[i].y > SCREEN_H || /*ships[i].y < 0 ||*/ ships[i].x > SCREEN_W || ships[i].x < 0) {
                for(int n=i; n<nb_ships-1; n++) {
                    ships[n] = ships[n+1];
                }
                nb_ships--;
                score.nb_dodge++;
            }
        }

        // Update missiles' coordinates and destroy lost ones.
        for(int i=nb_missiles-1; i>=0; i--) {
            if(missiles[i].speed < 1.0)
                missiles[i].speed += missile_models[missiles[i].type].acceleration;
            missiles[i].x += missiles[i].dx * missiles[i].speed;
            missiles[i].y += missiles[i].dy * missiles[i].speed + missiles[i].dy_shift + the_muship_gear.speed;
            if(missiles[i].y > SCREEN_H || missiles[i].y < 0 || missiles[i].x > SCREEN_W || missiles[i].x < 0) {
                for(int n=i; n<nb_missiles-1; n++) {
                    missiles[n] = missiles[n+1];
                }
                nb_missiles--;
            }
        }

        // Update particles' coordinates and regenerate lost ones.
        if(opti.particles) for(int i=0; i<MAX_PARTICLE; i++) {
            particles[i].y += particles[i].speed + the_muship_gear.speed;
            if(particles[i].y > SCREEN_H || particles[i].y < 0 || particles[i].x > SCREEN_W || particles[i].x < 0) {
                particles[i].type = rand() % particle_type_max;
                particles[i].speed = ((rand()%MAX_PARTICLE_SPEED)+1)*PARTICLE_SPEED_RATIO;
                particles[i].x = rand() % SCREEN_W;
                particles[i].y = 0;
            }
        }

        // Update explosions' state and destroy faded ones.
        if(opti.explosion) for(int i=nb_explosions-1; i>=0; i--) {
            explosions[i].y += the_muship_gear.speed; // Don't delete them when out of screen: they'll fade out.
            explosions[i].remains--;
            if(explosions[i].remains <= 0) {
                for(int n=i; n<nb_explosions-1; n++) {
                    explosions[n] = explosions[n+1];
                }
                nb_explosions--;
            }
        }

        // Spawn mechip accordingly to the run's definition.
        for(int i=0; i<run.nb_waves; i++) {
            if(run.waves[i].duration >= 0) {
                if(time >= run.waves[i].timestart) {
                    run.waves[i].duration -= the_muship_gear.speed;
                    run.waves[i].accumulator += run.waves[i].spawn_per_sec * the_muship_gear.speed;
                    if(run.waves[i].accumulator > 100) {
                        int x_spawn;
                        if(frontspawn <= 0) {
                            frontspawn = FRONTSPAWN_DELAY + rand()%FRONTSPAWN_VARIANCE;
                            x_spawn = the_muship_x;
                        } else {
                            frontspawn--;
                            x_spawn = rand() % SCREEN_W;
                        }
                        add_ship(&(ships[nb_ships++]), run.waves[i].model_id, x_spawn, -10, ship_models);
                        run.waves[i].accumulator -= 100;
                    }
                }
            }
        }

        // Fire mechile.
        for(int i=0; i<nb_ships; i++) {
            struct gear* gear = &(ship_models[ships[i].type].gear);
            for(int j=0; j<gear->nb_weapon; j++) {
                if(ships[i].cd[j] > 0) {
                    ships[i].cd[j]--;
                } else {
                    ships[i].cd[j] = gear->weapon[j].cd;
                }
                if(ships[i].cd[j] == gear->weapon[j].delay && nb_missiles < MAX_MISSILE-1) {
                    if(gear->weapon[j].missile == missile_mechautobullet) {
                        // Auto-aiming missile
                        float X = the_muship_x - ships[i].x;
                        float Y = the_muship_y - ships[i].y;
                        float H = sqrt(X*X+Y*Y);
                        if(H != 0) {
                            missiles[nb_missiles].dx = X/H * 2.0;
                            missiles[nb_missiles].dy = Y/H * 2.0; // Arbitrary speed of 2
                            missiles[nb_missiles].x = ships[i].x;
                            missiles[nb_missiles].y = ships[i].y;
                            missiles[nb_missiles].type = gear->weapon[j].missile;
                            missiles[nb_missiles].alliance = ALLIANCE_MECH;
                            missiles[nb_missiles].speed = 1.0;
                            nb_missiles++;
                        }
                    } else {
                        add_missile(&(missiles[nb_missiles++]), &(gear->weapon[j]), ships[i].x, ships[i].y, ALLIANCE_MECH, ship_models[ships[i].type].gear.speed);
                    }
                }
            }
        }

        // Fire muship weapons.
        if(!lose) {
            for(int i=0; i<the_muship_gear.nb_weapon; i++) {
                if(the_muship_cd[i] > 0) {
                    the_muship_cd[i]--;
                } else if(firekey) {
                    the_muship_cd[i] = the_muship_gear.weapon[i].cd;
                }
                if(firekey && the_muship_cd[i] == the_muship_gear.weapon[i].delay && nb_missiles < MAX_MISSILE-1) {
                    add_missile(&(missiles[nb_missiles++]), &(the_muship_gear.weapon[i]), the_muship_x, the_muship_y, ALLIANCE_MUSH, -the_muship_gear.speed);
                    // TODO : add the muship current Dy.
                    score.nb_shot++;
                }
            }
            for(int i=0; i<the_muship_gear.nb_special_actif; i++) {
                const struct special_actif* spe = &(the_muship_gear.special_actif[i]);
                if(the_muship_special_cd[i] > 0) {
                    the_muship_special_cd[i]--;
                } else if(spekey) {
                    the_muship_special_cd[i] = spe->cd;
                }
                if(the_muship_special_cd[i] < spe->cd - spe->delay
                    && the_muship_special_cd[i] > spe->cd - spe->delay - spe->duration
                ) {
                    switch(spe->type) {
                        case special_actif_beamlazer:
                            // Destroy any mechip in it's way.
                            for(int j=nb_ships-1; j>=0; j--) {
                                if(ships[j].x-SHIP_SIZE/2 < the_muship_x+spe->x+LAZER_WIDTH/2 &&
                                   ships[j].x+SHIP_SIZE/2 > the_muship_x+spe->x-LAZER_WIDTH/2 &&
                                   ships[j].y < the_muship_y)
                                {
                                    // Destroy the mechip.
                                    if(opti.explosion) add_explosion(explosions, &nb_explosions, ships[j].x, ships[j].y, EXPLOSION_TYPE_MECH);
                                    for(int n=j; n<nb_ships-1; n++) {
                                        ships[n] = ships[n+1];
                                    }
                                    nb_ships--;
                                    score.nb_destroy++;
                                }
                            }
                            break;
                    }
                }
            }
        }

        // Detect missile collisions.
        for(int i=0; i<nb_missiles; i++) {
            if(missiles[i].alliance == ALLIANCE_MUSH) {
                for(int j=0; j<nb_ships; j++) {
                    if(collision(missiles[i].x, missiles[i].y, missile_models[missiles[i].type].w, missile_models[missiles[i].type].h,
                                 ships[j].x, ships[j].y, SHIP_SIZE, SHIP_SIZE)) {
                        // Adding explosion even if ship isn't destroyed was originally a bug... But the effect seems cool enough to be kept.
                        if(opti.explosion) add_explosion(explosions, &nb_explosions, ships[j].x, ships[j].y, EXPLOSION_TYPE_MECH);
                        ships[j].hp -= missile_models[missiles[i].type].dammage;
                        // Destroy missile.
                        for(int n=i; n<nb_missiles-1; n++) {
                            missiles[n] = missiles[n+1];
                        }
                        nb_missiles--;
                        score.nb_hit++;
                        // Check if mechip is destroyed.
                        if(ships[j].hp <= 0) {
                            for(int n=j; n<nb_ships-1; n++) {
                                ships[n] = ships[n+1];
                            }
                            nb_ships--;
                            score.nb_destroy++;
                        }
                        break;
                    }
                }
            } else if(!lose && missiles[i].alliance == ALLIANCE_MECH) {
                if(collision(missiles[i].x, missiles[i].y, missile_models[missiles[i].type].w, missile_models[missiles[i].type].h,
                             the_muship_x, the_muship_y, the_muship_w, the_muship_h)) {
                    if(opti.explosion) add_explosion(explosions, &nb_explosions, the_muship_x, the_muship_y, EXPLOSION_TYPE_MUSH);
                    // Destroy missile.
                    for(int n=i; n<nb_missiles-1; n++) {
                        missiles[n] = missiles[n+1];
                    }
                    nb_missiles--;
                    the_muship_gear.hp--;
                    if(the_muship_gear.hp <= 0)
                        lose++;
                }
            }
        }

        // Detect inter ship collisions.
        if(!lose) {
            for(int i=0; i<nb_ships; i++) {
                if(abs(the_muship_x - ships[i].x) < 3*SHIP_SIZE && abs(the_muship_y - ships[i].y) < 3*SHIP_SIZE) {
                    if(!ships[i].skirt) {
                        if(!mute) {
                            add_sound(sound_thruster_doppler, 0.5);
                        }
                        ships[i].skirt = true;
                    }
                    if(collision(the_muship_x, the_muship_y, the_muship_w, the_muship_h, ships[i].x, ships[i].y, SHIP_SIZE, SHIP_SIZE)) {
                        int collision_power = 0;
                        for(int i=0; i<the_muship_gear.nb_special_passif; i++) {
                            if(the_muship_gear.special_passif[i].type == special_passif_rostrum) {
                                if(upkey) // Only count if muship is moving forward.
                                    collision_power += RAM_POWER;
                            } else if(the_muship_gear.special_passif[i].type == special_passif_spikedshield) {
                                collision_power += SPIKE_POWER;
                            }
                        }
                        ships[i].hp -= collision_power;
                        if(ships[i].hp > 0) { // If mech resisted ram attack, the muship is destroyed.
                            if(opti.explosion) add_explosion(explosions, &nb_explosions, the_muship_x, the_muship_y, EXPLOSION_TYPE_MUSH);
                            the_muship_gear.hp = 0;
                            lose = 1;
                        } // If mech was destroyed by ram attack, the muship survive.
                        // Anyway, destroy the mechip.
                        if(opti.explosion) add_explosion(explosions, &nb_explosions, ships[i].x, ships[i].y, EXPLOSION_TYPE_MECH);
                        for(int n=i; n<nb_ships-1; n++) {
                            ships[n] = ships[n+1];
                        }
                        nb_ships--;
                        score.nb_destroy++;
                    }
                }
            }
        }

        // Draw the screen.

        // Clear screen.
        SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
        SDL_RenderClear(renderer);

        // Draw particles.
        if(opti.particles) {
            dst.w = dst.h = PARTICLE_SIZE;
            for(int i=0; i<MAX_PARTICLE; i++) {
                // dst.x = particles[i].x - (PARTICLE_SIZE/2);
                // dst.y = particles[i].y - (PARTICLE_SIZE/2);
                // Particles are so small... There is no need to center them.
                dst.x = particles[i].x;
                dst.y = particles[i].y;
                SDL_RenderCopy(renderer, particle_texture[particles[i].type], NULL, &dst);
            }
        }

        // Draw aiming lazer.
        if(!lose) {
            for(int i=0; i<the_muship_gear.nb_special_passif; i++) {
                if(the_muship_gear.special_passif[i].type == special_passif_aiminglazer) {
                    struct special_passif* spe = &(the_muship_gear.special_passif[i]);
                    SDL_SetRenderDrawColor(renderer, 128, 0, 0, 255);
                    SDL_RenderDrawLine(renderer, the_muship_x+spe->x, the_muship_y+spe->y, the_muship_x+spe->x, 0);
                }
            }
        }

        // Draw mush lazers.
        if(!lose /* && spekey */) {
            for(int i=0; i<the_muship_gear.nb_special_actif; i++) {
                const struct special_actif* spe = &(the_muship_gear.special_actif[i]);
                if(the_muship_special_cd[i] > spe->cd - spe->delay) {
                    // Draw charging halo.
/*
                    SDL_SetRenderDrawColor(renderer, 0, 192, 255, 192);
                    for(int x=0; x<LAZER_HALO_RADIUS; x++) {
                        for(int y=0; y<LAZER_HALO_RADIUS; y++) {
                            if(rand()%2) {
                                SDL_RenderDrawPoint(renderer, the_muship_x+spe->x+x-LAZER_HALO_RADIUS/2, the_muship_y+spe->y+y-LAZER_HALO_RADIUS/2);
                            }
                        }
                    }
*/
                } else if(the_muship_special_cd[i] > spe->cd - spe->delay - spe->duration) {
                    // Draw lazer beam.
                    switch(spe->type) {
                        case special_actif_beamlazer:
                            for(int k=0; k<LAZER_WIDTH; k++) {
                                SDL_SetRenderDrawColor(renderer, 0, (5-abs(k-5))*50, 255, 192);
                                SDL_RenderDrawLine(renderer, the_muship_x + spe->x - LAZER_WIDTH/2 + k, 0,
                                                             the_muship_x + spe->x - LAZER_WIDTH/2 + k, the_muship_y + spe->y);
                            }
                            break;
                    }
                }
            }
        }

        // Draw mech lazers.
        for(int i=0; i<nb_ships; i++) {
            for(int j=0; j<ship_models[ships[i].type].gear.nb_special_actif; j++) {
                const struct special_actif* spe = &(ship_models[ships[i].type].gear.special_actif[j]);
                if(ships[i].special_cd[j] > 0) {
                    ships[i].special_cd[j]--;
                } else {
                    ships[i].special_cd[j] = spe->cd;
                }
                if(ships[i].special_cd[j] > spe->cd - spe->delay) {
                    // Draw charging halo.
                } else if(ships[i].special_cd[j] > spe->cd - spe->delay - spe->duration) {
                    // Draw lazer beam.
                    switch(spe->type) {
                        case special_actif_beamlazer:
                            for(int k=0; k<LAZER_WIDTH; k++) {
                                SDL_SetRenderDrawColor(renderer, 255, (5-abs(k-5))*50, 0, 192);
                                SDL_RenderDrawLine(renderer, ships[i].x + spe->x - LAZER_WIDTH/2 + k, SCREEN_H,
                                                             ships[i].x + spe->x - LAZER_WIDTH/2 + k, ships[i].y + spe->y);
                            }
                            if(!lose) {
                                // Detect collision with the muship.
                                if(the_muship_x-the_muship_w/2 < ships[i].x+spe->x+LAZER_WIDTH/2 &&
                                   the_muship_x+the_muship_w/2 > ships[i].x+spe->x-LAZER_WIDTH/2 &&
                                   the_muship_y > ships[i].y)
                                {
                                    if(opti.explosion) add_explosion(explosions, &nb_explosions, the_muship_x, the_muship_y, EXPLOSION_TYPE_MUSH);
                                    the_muship_gear.hp--;
                                    if(the_muship_gear.hp <= 0)
                                        lose++;
                                }
                            }
                            break;
                    }
                }
            }
        }

        // Draw missiles.
        for(int i=0; i<nb_missiles; i++) {
            dst.w = missile_models[missiles[i].type].w;
            dst.h = missile_models[missiles[i].type].h;
            dst.x = missiles[i].x - dst.w/2;
            dst.y = missiles[i].y - dst.h/2;
            if(opti.trails && missile_models[missiles[i].type].color_trail != NULL) {
                draw_trail(renderer, missiles[i].x, missiles[i].y - dst.w/2, 5, -5*missiles[i].dy*missiles[i].speed, missile_models[missiles[i].type].color_trail);
            }
            SDL_RenderCopy(renderer, missile_models[missiles[i].type].texture, NULL, &dst);
        }

        // Draw mechips.
        dst.w = dst.h = SHIP_SIZE;
        for(int i=0; i<nb_ships; i++) {
            if(opti.trails) for(int j=0; j<ship_models[ships[i].type].gear.nb_thruster; j++) {
                struct thruster* thruster = &(ship_models[ships[i].type].gear.thruster[j]);
                draw_trail(renderer, thruster->x + ships[i].x, thruster->y + ships[i].y, thruster->w, thruster->h, thruster->color_trail);
            }
            dst.x = ships[i].x - (SHIP_SIZE/2);
            dst.y = ships[i].y - (SHIP_SIZE/2);
            SDL_RenderCopy(renderer, ship_models[ships[i].type].texture, NULL, &dst);
        }

        // Draw the muship.
        if(!lose) {
            dst.w = the_muship_w;
            dst.h = the_muship_h;
            dst.x = the_muship_x - (the_muship_w/2);
            dst.y = the_muship_y - (the_muship_h/2);

            // Draw thrusters' trails.
            if(opti.trails) {
                int accelerate = false;
                int brake = false;
                if(upkey && !dwnkey) {
                    accelerate = true;
                } else if(dwnkey && !upkey) {
                    brake = true;
                }
                for(int i=0; i<the_muship_gear.nb_thruster; i++) {
                    int length = the_muship_gear.thruster[i].h;
                    if(accelerate) {
                        length *= 1.5;
                    } else if(brake) {
                        length /= 2;
                    }
                    draw_trail(renderer,
                        the_muship_x + the_muship_gear.thruster[i].x+1,
                        the_muship_y + the_muship_gear.thruster[i].y,
                        the_muship_gear.thruster[i].w,
                        length,
                        the_muship_gear.thruster[i].color_trail
                    );
                }
            }

            // Draw the muship.
            SDL_RenderCopy(renderer, the_muship_texture, NULL, &dst);
        }

        // Draw explosions.
        if(opti.explosion) {
            dst.w = dst.h = SHIP_SIZE;
            for(int i=0; i<nb_explosions; i++) {
                dst.x = explosions[i].x - (SHIP_SIZE/2);
                dst.y = explosions[i].y - (SHIP_SIZE/2);
                SDL_RenderCopy(renderer, explosion_texture[phase(explosions[i].remains)][explosions[i].type], NULL, &dst);
            }
        }

        // Draw the dialogues accordingly to the run description.
        for(int i=0; i<run.nb_dialogue; i++) {
            if(time >= run.dialogues[i].timestart && run.dialogues[i].duration > 0) {
                run.dialogues[i].duration--;
                dst.w = run.dialogues[i].w;
                dst.h = run.dialogues[i].h;
                dst.x = 30;
                dst.y = 30 + run.dialogues[i].y_shift;
                SDL_RenderCopy(renderer, run.dialogues[i].texture, NULL, &dst);
            }
        }

        // Draw hitpoints.
        icon_shield_dst.y = SCREEN_H - 10 - icon_shield_dst.h;
        for(int i=0; i<the_muship_gear.hp; i++) {
            icon_shield_dst.x = 10 + (10+icon_shield_dst.w)*i;
            SDL_RenderCopy(renderer, icon_shield, NULL, &icon_shield_dst);
        }

        // Display the frame!!
        SDL_RenderPresent(renderer);

        // Check ends of game.

        if(lose) {
            lose++;
        } else if(!win) {

            if(nb_ships > 0)
                goto no_win_now;
            for(int i=0; i<run.nb_waves; i++)
                if(run.waves[i].timestart+run.waves[i].duration > time)
                    goto no_win_now;
            for(int i=0; i<run.nb_dialogue; i++)
                if(run.dialogues[i].timestart+run.dialogues[i].duration > time)
                    goto no_win_now;
            for(int i=0; i<nb_missiles; i++)
                if(missiles[i].alliance != ALLIANCE_MUSH)
                    goto no_win_now;

            // If we reached it, then the run is totally over : we can claim victory.
            win = 1; // First frame of victory.
no_win_now:
            win = win;
        } else {
            win++; // One more frame of victory.
        }

        if(lose > ENDGAME_DELAY) {
            stop = 1;
            result = false;
        }

        if(win > ENDGAME_DELAY && !lose) {
            stop = 1;
            result = true;
        }

        // Delay until next frame.
        // time++;
        time += the_muship_gear.speed;
        score.time++;

        int sleeptime = last_tick + 10 - SDL_GetTicks();
        if(sleeptime > 0) {
            SDL_Delay(sleeptime);
        }
        last_tick = SDL_GetTicks();
    }

    // Clean up.
    SDL_DestroyTexture(the_muship_texture);
    SDL_DestroyTexture(icon_shield);
    unload_mechip_models(nb_ship_models, ship_models);
    if(!mute) {
        stop_sound();
    }
    for(int i=0; i<missile_type_max; i++) {
        SDL_DestroyTexture(missile_models[i].texture);
    }
    if(opti.particles) for(int i=0; i<particle_type_max; i++) {
        SDL_DestroyTexture(particle_texture[i]);
    }
    if(opti.explosion) for(int i=0; i<EXPLOSION_PHASE_MAX; i++) {
        for(int j=0; j<EXPLOSION_TYPE_MAX; j++) {
            SDL_DestroyTexture(explosion_texture[i][j]);
        }
    }
    for(int i=0; i<run.nb_dialogue; i++) {
        SDL_DestroyTexture(run.dialogues[i].texture);
    }

    // Return.
    if(result >= 0) {
        gameover(renderer, score, result);
    }
    return(result);
}
