
/* Copyright  2014 Pierre Vidal <p.terkal@gmail.com>
 * This program is free software. It comes without no warranty.
 * You can redistribute it and/or modify it under the terms of the
 * Do What The Fuck You Want To Public License, Version 2, as published by Sam Hocevar.
 * See the COPYING file or http://www.wtfpl.net/ for more details. */

#include "main.h"

#include <SDL2/SDL_ttf.h>
#include <stdio.h>
#include <time.h>
#include <string.h> // strlen() in error()

TTF_Font* font;

SDL_Texture* loadTexture(SDL_Renderer* renderer, char* filename) {
    SDL_Texture* returnme;
    SDL_Surface* surface = SDL_LoadBMP(filename);
    if(surface == NULL) {
        sdl_error("Unable to load bitmap");
        return(NULL);
    }
    returnme = SDL_CreateTextureFromSurface(renderer, surface);
    SDL_FreeSurface(surface);
    return(returnme);
}

SDL_Texture* loadTextureSize(SDL_Renderer* renderer, char* filename, int* w, int* h) {
    SDL_Texture* returnme;
    SDL_Surface* surface = SDL_LoadBMP(filename);
    if(surface == NULL) {
        sdl_error("Unable to load bitmap");
        return(NULL);
    }
    returnme = SDL_CreateTextureFromSurface(renderer, surface);
    *w = surface->w;
    *h = surface->h;
    SDL_FreeSurface(surface);
    return(returnme);
}

SDL_Texture* loadTextureText(SDL_Renderer* renderer, char* text, int* w, int * h, SDL_Color const * const color) {
    SDL_Surface* surface;
    SDL_Texture* texture;

    surface = TTF_RenderText_Solid(font, text, *color);
    if(surface == NULL) {
        sdl_error("Unable to render text");
        return(NULL);
    }
    *w = surface->w;
    *h = surface->h;
    texture = SDL_CreateTextureFromSurface(renderer, surface);
    SDL_FreeSurface(surface);
    return(texture);
}

void error(char* msg) {
    char buf[strlen(msg)+8];
    sprintf(buf, "ERROR: %s", msg);
    perror(buf);
}

void sdl_error(char* msg) {
    printf("ERROR: %s: %s\n", msg, SDL_GetError());
}

void generate_stars(struct star* stars, int nb_stars) {
    for(int i=0; i<nb_stars; i++) {
        stars[i].x = rand()%SCREEN_W;
        stars[i].y = rand()%SCREEN_H;
        stars[i].r = rand()%256;
        stars[i].g = rand()%256;
        stars[i].b = rand()%256;
    }
}

void draw_stars(SDL_Renderer* renderer, struct star* stars, int nb_stars) {
    for(int i=0; i<nb_stars; i++) {
        SDL_SetRenderDrawColor(renderer, stars[i].r, stars[i].g, stars[i].b, 255);
        SDL_RenderDrawPoint(renderer, stars[i].x, stars[i].y);
    }
}

int WinMain() {
    return(main(0, NULL));
}

int main(int argc, char** argv) {
    SDL_Renderer* renderer;
    SDL_Window* screen;

    // Redirect output to console.
    freopen("CON", "w", stdout);
    freopen("CON", "w", stderr);

    // Init random generator.
    srand(time(NULL));

    // Initialize SDL video and audio.
    if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0) {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        return(1);
    }
    atexit(SDL_Quit);

    // Initialize SDL TTF.
    if(TTF_Init() == -1) {
        printf("Unable to init TTF : %s\n", TTF_GetError());
        return(1);
    }
    atexit(TTF_Quit);

    // Load font.
    font = TTF_OpenFont(path("", "visitor1.ttf"), 20);
    if(font == NULL) {
        printf("Unable to load font : %s\n", TTF_GetError());
        return(1);
    }
    lineskip = TTF_FontLineSkip(font);

    // Create window and renderer.
    screen = SDL_CreateWindow(NULL, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 0, 0, 0);
    if(!screen) {
        printf("Unable to open new window : %s\n", SDL_GetError());
        return(1);
    }
    renderer = SDL_CreateRenderer(screen, -1, 0);
    if(!renderer) {
        printf("Unable to create new renderer : %s\n", SDL_GetError());
        return(1);
    }

    // Configure window.

    SDL_SetWindowTitle(screen, "SHMUP : Hidden Mushroom Ultra Portal");
    // SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");  // make the scaled rendering look smoother. // DON'T DO THAT!!!
    SDL_RenderSetLogicalSize(renderer, SCREEN_W, SCREEN_H);
    SDL_SetWindowSize(screen, SCREEN_W, SCREEN_H);
    SDL_SetWindowPosition(screen, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
    SDL_ShowCursor(SDL_DISABLE);

    SDL_Surface* icon = SDL_LoadBMP(path("", "icon.bmp"));
    if(!icon) {
        sdl_error("Unable to load icon");
    } else {
        SDL_SetWindowIcon(screen, icon);
    }

    load_local(DEFAULT_LANGUAGE);

    opti.explosion = true;
    opti.particles = true;
    opti.trails = true;
    global_volume = 1.0;
    mute = false;

    // Launch the game through the menu.
    menu(renderer, screen);

    // Clean up.
    unload_local();
    SDL_DestroyWindow(screen);
    // SDL_DestroyRenderer(renderer); // FIXME : Why does it bug ?!?
    // SDL_Quit(); // Already done with atexit().
    TTF_CloseFont(font);

    return(0);
}
